import { ABView }  from "./cardbookHTMLCardsView.js";
import { cardbookNewPreferences } from "./preferences/cardbookNewPreferences.mjs";
import { cardbookHTMLUtils } from "./cardbookHTMLUtils.mjs";

export function CBTree(dirPrefId, tablename, availableColumns, events) {
	this.columns = [];
	this.sortResource = "";
	this.sortDirection = "";
	this.availableColumns = [];

	this.deleteKey = events.deleteKey || null;
	this.enterKey = events.enterKey || null;
	this.dblclick = events.dblclick || null;
	this.dragstart = events.dragstart || null;
	this.dragover = events.dragover || null;
	this.drop = events.drop || null;

	this.dirPrefId = dirPrefId;

	this.cardsList = document.getElementById(tablename);
	if (availableColumns) {
		this.availableColumns = availableColumns;
	}

	this.init();
}
CBTree.prototype = {
	init() {
		this.table = this.cardsList.table;
		this.table.editable = true;
		this.table.setBodyID(`${this.cardsList.id}Body`);
		this.cardsList.setAttribute("rows", "cb-table-card-row");

		this.table.addEventListener("columns-changed", this);
		this.table.addEventListener("sort-changed", this);
		this.table.addEventListener("column-resized", this);
		this.table.addEventListener("reorder-columns", this);
		this.table.addEventListener("restore-columns", this);
		this.cardsList.addEventListener("keydown", this);
		this.cardsList.addEventListener("dblclick", this);
		this.cardsList.addEventListener("dragstart", this);
		this.cardsList.addEventListener("dragover", this);
		this.cardsList.addEventListener("drop", this);

		window.addEventListener("uidensitychange", () => this.densityChange());
		customElements.whenDefined("cb-table-card-row").then(() => this.densityChange());
	},

	handleEvent(event) {
		switch (event.type) {
			case "keydown":
				this._onKeyDown(event);
				break;
			case "dblclick":
				if (this.dblclick) {
					this.dblclick(event);
				}
				break;
			case "dragstart":
				if (this.dragstart) {
					this.dragstart(event);
				}
				break;
			case "dragover":
				if (this.dragover) {
					this.dragover(event);
				}
				break;
			case "drop":
				if (this.drop) {
					this.drop(event);
				}
				break;
			case "columns-changed":
				this._onColumnsChanged(event.detail);
				break;
			case "sort-changed":
				this._onSortChanged(event);
				break;
			case "column-resized":
				this._onColumnResized(event);
				break;
			case "reorder-columns":
				this._onReorderColumns(event.detail);
				break;
			case "restore-columns":
				this._onRestoreDefaultColumns();
				break;
		}
	},

	_onKeyDown(event) {
		if (event.altKey || event.shiftKey) {
			return;
		}
		let modifier = event.ctrlKey;
		let antiModifier = event.metaKey;
		if (navigator.userAgent.indexOf("Mac") !== -1) {
			[modifier, antiModifier] = [antiModifier, modifier];
		}
		if (antiModifier) {
			return;
		}
		switch (event.key) {
			case "a":
				if (modifier) {
					this.cardsList.view.selection.selectAll();
					this.cardsList.dispatchEvent(new CustomEvent("select"));
					event.preventDefault();
				}
				break;
			case "Delete":
				if (!modifier) {
					if (this.deleteKey) {
						this.deleteKey(event);
					}
					event.preventDefault();
				}
				break;
			case "Enter":
				if (!modifier) {
					if (this.enterKey) {
						this.enterKey(event);
					}
					event.preventDefault();
				}
				break;
		}
	},

	_onRestoreDefaultColumns() {
		let defaultColumns = cardbookNewPreferences.getDefaultTreeColumns(this.cardsList.id).split(",");
		defaultColumns = defaultColumns.map(x => this.cardsList.id + x.split(":")[0]);

		function arraymove(arr, from, to) {
			var element = arr[from];
			arr.splice(from, 1);
			arr.splice(to, 0, element);
		}
		let index = 0;
		for (let column of defaultColumns) {
			let from = this.columns.map(element => element.id).indexOf(column);
			arraymove(this.columns, from, index);
			index++;
		}
		let ordinal = 0
		for (let column of this.columns) {
			column.hidden = !defaultColumns.includes(column.id);
			column.ordinal = ordinal++;
		}

		this.table.setColumns(this.columns);
		this.table.querySelector(".sorting")?.classList.remove("sorting", "ascending", "descending");
		this.table.querySelector(`#${this.cardsList.id}${this.sortResource} button`)?.classList.add("sorting", this.sortDirection);
		this.cardsList.invalidate();
		this.setColumnsWidths(this.cardsList.id, this.table.getColumnsWidths());
	},

	_onColumnResized(event) {
		this.setColumnsWidths(this.cardsList.id, this.table.getColumnsWidths());
	},

	async _onSortChanged(event) {
		if (this.cardsList.view) {
			const { sortColumn, sortDirection } = this.cardsList.view;
			const column = event.detail.column;
			await this.sortRows(column, sortColumn == column && sortDirection == "ascending" ? "descending" : "ascending");
		}
	},

	_onColumnsChanged(data) {
		let column = data.value;
		let checked = data.target.classList.contains("active");
		let visibleOrdinal = this.columns.filter(x => x.hidden === false).map(x => x.ordinal);
		let valuesVisibleOrdinal = Object.values(visibleOrdinal);
		let maxVisibleOrdinal = Math.max(...valuesVisibleOrdinal);

		for (let columnDef of this.columns) {
			if (columnDef.id == column) {
				columnDef.hidden = !checked;
				columnDef.ordinal = maxVisibleOrdinal++;
				break;
			}
		}
		this.table.updateColumns(this.columns);
		this.cardsList.invalidate();
		this.setColumnsWidths(this.cardsList.id, this.table.getColumnsWidths());
	},

	_onReorderColumns(data) {
		this.columns = data.columns;
		this.table.updateColumns(this.columns);
		this.cardsList.invalidate();
		this.setColumnsWidths(this.cardsList.id, this.table.getColumnsWidths());
	},

	setColumnsWidths(tableId, columnsToSave) {
		let regexp = new RegExp(`^${this.cardsList.id}`);
		columnsToSave = columnsToSave.map(x => x.replace(regexp, ""));
		cardbookHTMLUtils.setPrefValue(`${tableId}Columns`, columnsToSave.join(","));
	},

	async setHeader() {
		let treeColumns = cardbookNewPreferences.getTreeColumns(this.cardsList.id);
		this.sortResource = cardbookHTMLUtils.getPrefValue(`${this.cardsList.id}SortResource`);
		this.sortResource = `${this.cardsList.id}${this.sortResource}`
		this.sortDirection = cardbookHTMLUtils.getPrefValue(`${this.cardsList.id}SortDirection`);
		this.columns = [];
		let savedColumns = {};
		let ordinal = 0;
		for (let col of treeColumns.split(",")) {
			let colName = col.split(":")[0];
			let colId = `${this.cardsList.id}${colName}`;
			let colWidth = typeof col.split(":")[1] !== 'undefined' ? col.split(":")[1] : 0;
			savedColumns[colId] = {};
			savedColumns[colId].width = colWidth;
			savedColumns[colId].ordinal = ordinal++;
		}

		let allColumns = [];
		if (this.availableColumns.length) {
			allColumns = JSON.parse(JSON.stringify(this.availableColumns));
		} else {
			allColumns = await messenger.runtime.sendMessage({query: "cardbook.getAllAvailableColumns", mode: "cardstree"});
		}
		for (let column of allColumns) {
			let col = {};
			col.id = `${this.cardsList.id}${column[0]}`;
			col.hidden = typeof savedColumns[col.id] !== 'undefined' ? false : true;
			col.width = typeof savedColumns[col.id] !== 'undefined' ? savedColumns[col.id].width : 0;
			col.ordinal = typeof savedColumns[col.id] !== 'undefined' ? savedColumns[col.id].ordinal : allColumns.length + 1;
			col.label = column[1];
			this.columns.push(col);
		}
		this.columns.sort(function(a, b) {
			return a.ordinal-b.ordinal
		})
		this.table.setColumns(this.columns);
		this.table.restoreColumnsWidths(savedColumns);
	},

	async displayBook(cards) {
		this.cardsList.view = new ABView(cards, this.cardsList.id, this.sortResource, this.sortDirection);
		await this.sortRows(this.sortResource, this.sortDirection);
	},

	async sortRows(column, direction) {
		// Unmark the header of previously sorted column, then mark the header of
		// the column to be sorted.
		this.table.querySelector(".sorting")?.classList.remove("sorting", "ascending", "descending");
		this.table.querySelector(`#${column} button`)?.classList.add("sorting", direction);
		if (this.cardsList.view.sortColumn == column && this.cardsList.view.sortDirection == direction) {
			return;
		}

		this.cardsList.view.sortBy(column, direction);

		let regexp = new RegExp(`^${this.cardsList.id}`);
		let sortResource = column.replace(regexp, "");
		await cardbookHTMLUtils.setPrefValue(`${this.cardsList.id}SortResource`, sortResource);
		await cardbookHTMLUtils.setPrefValue(`${this.cardsList.id}SortDirection`, direction);
	},

	async densityChange() {
		let tableRowClass = customElements.get("cb-table-card-row");
		let density = await messenger.runtime.sendMessage({ query: "cardbook.getDensity"});
		switch (density) {
			case 0: // MODE_COMPACT
				tableRowClass.ROW_HEIGHT = 18;
				break;
			case 2: // MODE_TOUCH
				tableRowClass.ROW_HEIGHT = 32;
				break;
			default: // MODE_NORMAL
				tableRowClass.ROW_HEIGHT = 22;
				break;
		}
		this.cardsList.invalidate();
	},
};
